"""Implementation of generic SHV RPC broker."""
import argparse
import asyncio
import configparser
import logging

from .rpcbroker import RpcBroker
from .rpcbrokerconfig import RpcBrokerConfig

logger = logging.getLogger(__name__)
log_levels = (
    logging.DEBUG,
    logging.INFO,
    logging.WARNING,
    logging.ERROR,
    logging.CRITICAL,
)


def parse_args():
    """Parse passed arguments and return result."""
    parser = argparse.ArgumentParser("pyshvbroker", description="Silicon Heaven broker")
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {RpcBroker.Client.APP_VERSION}",
    )
    parser.add_argument(
        "-v",
        action="count",
        default=0,
        help="Increase verbosity level of logging",
    )
    parser.add_argument(
        "-q",
        action="count",
        default=0,
        help="Decrease verbosity level of logging",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        default="",
        help="Configuration file",
    )
    return parser.parse_args()


async def main() -> None:
    """Application's entrypoint."""
    args = parse_args()

    logging.basicConfig(
        level=log_levels[sorted([1 - args.v + args.q, 0, len(log_levels) - 1])[1]],
        format="[%(asctime)s] [%(levelname)s] - %(message)s",
    )

    config = configparser.ConfigParser()
    if args.config:
        config.read(args.config)
    brokerconf = RpcBrokerConfig.load(config)
    broker = RpcBroker(brokerconf)
    await broker.serve_forever()


if __name__ == "__main__":
    asyncio.run(main())



"""Configuration of the broker."""
import configparser
import dataclasses
import hashlib
import typing

from ..rpcmethod import RpcMethodAccess
from ..rpcurl import RpcLoginType, RpcUrl


class RpcBrokerConfig:
    """Generic store of broker configuration."""

    @dataclasses.dataclass(frozen=True)
    class Rule:
        """Description of access rule."""

        name: str
        """Name of the rule."""
        path: str = ""
        """Path prefix that need to match for this rule to apply."""
        methods: frozenset[str] = dataclasses.field(default_factory=frozenset)
        """Set of methods this rule matches. Empty set matches all methods."""

        def applies(self, path: str, method: str) -> bool:
            """Check if this rule applies."""
            return (not self.methods or method in self.methods) and path.startswith(
                self.path
            )

    @dataclasses.dataclass(frozen=True)
    class Role:
        """Generic roles assigned to the users."""

        name: str
        """Name of the role."""
        access: RpcMethodAccess = RpcMethodAccess.BROWSE
        """Access level granted to the user by this role."""
        rules: frozenset["RpcBrokerConfig.Rule"] = dataclasses.field(
            default_factory=frozenset
        )
        """Rules used to check if this role should apply."""
        roles: frozenset["RpcBrokerConfig.Role"] = dataclasses.field(
            default_factory=frozenset
        )
        """Additional roles to check."""

        def applies(self, path: str, method: str) -> bool:
            """Check if this role applies by this rules."""
            return any(rule.applies(path, method) for rule in self.rules)

    @dataclasses.dataclass(frozen=True)
    class User:
        """User's login information."""

        name: str
        password: str
        login_type: RpcLoginType = RpcLoginType.SHA1
        roles: frozenset["RpcBrokerConfig.Role"] = dataclasses.field(
            default_factory=frozenset
        )

        def all_roles(self) -> typing.Iterator["RpcBrokerConfig.Role"]:
            """Iterate over all roles assigned to this user."""
            rlst = list(self.roles)
            while rlst:
                role = rlst.pop()
                yield role
                rlst.extend(role.roles)

        def access_level(self, path: str, method: str) -> RpcMethodAccess | None:
            """Deduce access level (if any) for this user on given path and method."""
            for role in self.all_roles():
                if role.applies(path, method):
                    return role.access
            return None

        @property
        def shapass(self) -> str:
            """Password in SHA1 format."""
            if self.login_type is RpcLoginType.SHA1:
                return self.password
            m = hashlib.sha1()
            m.update(self.password.encode("utf-8"))
            return m.hexdigest()

        def validate_password(
            self,
            password: str,
            nonce: str,
            login_type: RpcLoginType = RpcLoginType.SHA1,
        ) -> bool:
            """Check if given password is correct."""
            if login_type is RpcLoginType.PLAIN:
                return self.password == password
            if login_type is RpcLoginType.SHA1:
                m = hashlib.sha1()
                m.update(nonce.encode("utf-8"))
                m.update(self.shapass.encode("utf-8"))
                return m.hexdigest() == password
            return False

    def __init__(self) -> None:
        self.listen: dict[str, RpcUrl] = {}
        """URLs the broker should listen on."""
        self._users: dict[str, "RpcBrokerConfig.User"] = {}
        self._roles: dict[str, "RpcBrokerConfig.Role"] = {}
        self._rules: dict[str, "RpcBrokerConfig.Rule"] = {}

    def add_user(self, info: User) -> None:
        """Add or replace user.

        :param info: User description.
        :raise ValueError: in case it specifies unknown role.
        """
        for role in info.roles:
            if role is not self._roles.get(role.name, None):
                raise ValueError(f"Invalid role '{role.name}'")
        self._users[info.name] = info

    def user(self, name: str) -> User:
        """Get user for given name.

        :param name: Name of the user.
        :raise KeyError: when there is no such user.
        """
        return self._users[name]

    def users(self) -> typing.Iterator[User]:
        """Iterate over all users."""
        yield from self._users.values()

    def add_role(self, role: Role) -> None:
        """Add or replace role.

        :param info: role description.
        :raise ValueError: in case it specifies unknown rule.
        """
        for rule in role.rules:
            if rule is not self._rules.get(rule.name, None):
                raise ValueError(f"Invalid rule '{rule.name}'")
        for srole in role.roles:
            if srole is not self._roles.get(srole.name, None):
                raise ValueError(f"Invalid role '{srole.name}'")
        oldrole = self._roles.get(role.name, None)
        if oldrole is not None:
            for name, srole in self._roles.items():
                if oldrole in srole.roles:
                    self._roles[name] = dataclasses.replace(
                        srole, roles=(srole.roles ^ {oldrole, role})
                    )
            for name, user in self._users.items():
                if oldrole in user.roles:
                    self._users[name] = dataclasses.replace(
                        user, roles=(user.roles ^ {oldrole, role})
                    )
        self._roles[role.name] = role

    def role(self, name: str) -> Role:
        """Get role with given name.

        :param name: Name of the role.
        :raise KeyError: when there is no such role.
        """
        return self._roles[name]

    def roles(self) -> typing.Iterator[Role]:
        """Iterate over all roles."""
        yield from self._roles.values()

    def add_rule(self, rule: Rule) -> None:
        """Add or replace rule.

        :param info: rule description.
        """
        oldrule = self._rules.get(rule.name, None)
        if oldrule is not None:
            for name, role in self._roles.items():
                if oldrule in role.rules:
                    self._roles[name] = dataclasses.replace(
                        role, rules=(role.rules ^ {oldrule, rule})
                    )
        self._rules[rule.name] = rule

    def rule(self, name: str) -> Rule:
        """Get rule with given name.

        :param name: Name of the rule.
        :raise KeyError: when there is no such rule.
        """
        return self._rules[name]

    def rules(self) -> typing.Iterator[Rule]:
        """Iterate over all rules."""
        yield from self._rules.values()

    def login(
        self, user: str, password: str, nonce: str, login_type: RpcLoginType
    ) -> typing.Optional["RpcBrokerConfig.User"]:
        """Check the login and provide user if login is correct."""
        if user not in self._users or not self._users[user].validate_password(
            password, nonce, login_type
        ):
            return None
        return self.user(user)

    @classmethod
    def load(cls, config: configparser.ConfigParser) -> "RpcBrokerConfig":
        """Load configuration from ConfigParser."""
        res = cls()
        if "listen" in config:
            for name, url in config["listen"].items():
                res.listen[name] = RpcUrl.parse(url)
        for secname, sec in filter(lambda v: v[0].startswith("rules."), config.items()):
            res.add_rule(
                cls.Rule(
                    secname[6:],
                    sec.get("path", ""),
                    frozenset(sec.get("methods", "").split()),
                )
            )
        for secname, sec in filter(lambda v: v[0].startswith("roles."), config.items()):
            res.add_role(
                cls.Role(
                    secname[6:],
                    RpcMethodAccess.fromstr(sec.get("access", "")),
                    frozenset(
                        res.rule(rule_name)
                        for rule_name in sec.get("rules", "").split()
                    ),
                )
            )
        for secname, sec in filter(lambda v: v[0].startswith("roles."), config.items()):
            roles = sec.get("roles", "").split()
            name = secname[6:]
            if roles:
                res.add_role(
                    dataclasses.replace(
                        res.role(name),
                        roles=frozenset(map(res.role, roles)),
                    )
                )
        for secname, sec in filter(lambda v: v[0].startswith("users."), config.items()):
            if "password" in sec:
                password = sec["password"]
                login_type = RpcLoginType.PLAIN
            else:
                password = sec.get("sha1pass", "")
                login_type = RpcLoginType.SHA1
            res.add_user(
                cls.User(
                    secname[6:],
                    password,
                    login_type,
                    frozenset(
                        res.role(role_name)
                        for role_name in sec.get("roles", "").split()
                    ),
                )
            )
        return res



"""Check that we correctly parse and serialize our URL format."""
import pytest

from shv import RpcLoginType, RpcProtocol, RpcUrl

DATA = [
    ("localsocket:/dev/null", RpcUrl("/dev/null", protocol=RpcProtocol.LOCAL_SOCKET)),
    ("tcp://test@localhost:4242", RpcUrl("localhost", username="test", port=4242)),
    (
        "tcp://localhost:4242?devid=foo&devmount=/dev/null",
        RpcUrl("localhost", port=4242, device_id="foo", device_mount_point="/dev/null"),
    ),
    (
        "tcp://localhost:4242?devid=foo&devmount=/dev/null&password=test",
        RpcUrl(
            "localhost",
            port=4242,
            password="test",
            login_type=RpcLoginType.PLAIN,
            device_id="foo",
            device_mount_point="/dev/null",
        ),
    ),
    (
        "tcp://localhost:4242?devid=foo&devmount=/dev/null&shapass=xxxxxxxx",
        RpcUrl(
            "localhost",
            port=4242,
            password="xxxxxxxx",
            login_type=RpcLoginType.SHA1,
            device_id="foo",
            device_mount_point="/dev/null",
        ),
    ),
    ("tcp://[::]:4242", RpcUrl("::", port=4242)),
]


@pytest.mark.parametrize(
    "url,rpcurl",
    DATA
    + [
        ("", RpcUrl("", protocol=RpcProtocol.LOCAL_SOCKET)),
        ("/dev/null", RpcUrl("/dev/null", protocol=RpcProtocol.LOCAL_SOCKET)),
        ("//dev/null", RpcUrl("/dev/null", protocol=RpcProtocol.LOCAL_SOCKET)),
        (
            "localsocket://dev/null",
            RpcUrl("/dev/null", protocol=RpcProtocol.LOCAL_SOCKET),
        ),
        ("tcp://localhost", RpcUrl("localhost")),
        ("tcp://test@localhost:4242", RpcUrl("localhost", username="test", port=4242)),
        ("tcp://localhost?devid=foo", RpcUrl("localhost", device_id="foo")),
    ],
)
def test_parse(url, rpcurl):
    """Check that we correctly parse these URLs."""
    assert RpcUrl.parse(url) == rpcurl


def test_invalid():
    with pytest.raises(ValueError):
        RpcUrl.parse("foo://some")


def test_invalid_query():
    with pytest.raises(ValueError):
        RpcUrl.parse("tcp:localhost?missing=some")


@pytest.mark.parametrize("url,rpcurl", DATA)
def test_to_url(url, rpcurl):
    """Check that we correctly parse these URLs."""
    assert rpcurl.to_url() == url



"""Python implementation of Silicon Heaven."""
from . import chainpack, cpon
from .chainpack import ChainPackReader, ChainPackWriter
from .cpon import CponReader, CponWriter
from .rpcclient import RpcClient
from .rpcerrors import (
    RpcError,
    RpcErrorCode,
    RpcInternalError,
    RpcInvalidParamsError,
    RpcInvalidRequestError,
    RpcMethodCallCancelledError,
    RpcMethodCallExceptionError,
    RpcMethodCallTimeoutError,
    RpcMethodNotFoundError,
    RpcParseError,
)
from .rpcmessage import RpcMessage
from .rpcmethod import (
    RpcMethodAccess,
    RpcMethodDesc,
    RpcMethodFlags,
    RpcMethodSignature,
)
from .rpcserver import create_rpc_server
from .rpcurl import RpcLoginType, RpcProtocol, RpcUrl
from .simpleclient import SimpleClient
from .value import (
    SHVBool,
    SHVBytes,
    SHVDatetime,
    SHVDecimal,
    SHVDict,
    SHVFloat,
    SHVInt,
    SHVList,
    SHVMeta,
    SHVMetaType,
    SHVNull,
    SHVStr,
    SHVType,
    SHVUInt,
    is_shvbool,
    is_shvimap,
    is_shvmap,
    is_shvnull,
    shvget,
    shvmeta,
    shvmeta_eq,
)
from .valueclient import ValueClient

__all__ = [
    # cpon
    "cpon",
    "CponReader",
    "CponWriter",
    # chainpack
    "chainpack",
    "ChainPackReader",
    "ChainPackWriter",
    # rpcclient
    "RpcClient",
    # rpcserver
    "create_rpc_server",
    # rpcurl
    "RpcProtocol",
    "RpcLoginType",
    "RpcUrl",
    # rpcmessage
    "RpcMessage",
    # rpcmethod
    "RpcMethodSignature",
    "RpcMethodFlags",
    "RpcMethodAccess",
    "RpcMethodDesc",
    # rpcerror
    "RpcErrorCode",
    "RpcError",
    "RpcInternalError",
    "RpcInvalidParamsError",
    "RpcInvalidRequestError",
    "RpcMethodCallCancelledError",
    "RpcMethodCallExceptionError",
    "RpcMethodCallTimeoutError",
    "RpcMethodNotFoundError",
    "RpcParseError",
    # simpleclient
    "SimpleClient",
    # valueclient
    "ValueClient",
    # value
    "SHVMeta",
    "shvmeta",
    "shvmeta_eq",
    "SHVNull",
    "is_shvnull",
    "SHVBool",
    "is_shvbool",
    "SHVInt",
    "SHVUInt",
    "SHVFloat",
    "SHVDecimal",
    "SHVStr",
    "SHVBytes",
    "SHVDatetime",
    "SHVList",
    "SHVDict",
    "is_shvmap",
    "is_shvimap",
    "shvget",
    "SHVType",
    "SHVMetaType",
]



"""Chainpack data format reader and writer."""
import collections.abc
import datetime
import decimal
import math
import struct

from . import commonpack
from .value import SHVMeta, SHVMetaType, SHVType, SHVUInt


class ChainPack:
    """ChainPack constans and definitions."""

    ProtocolType = 1

    CP_Null = 128
    CP_UInt = 129
    CP_Int = 130
    CP_Double = 131
    CP_Bool = 132
    CP_Blob = 133
    CP_String = 134  # utf8 encoded string
    CP_List = 136
    CP_Map = 137
    CP_IMap = 138
    CP_MetaMap = 139
    CP_Decimal = 140
    CP_DateTime = 141
    CP_CString = 142
    CP_FALSE = 253
    CP_TRUE = 254
    CP_TERM = 255

    # UTC msec since 2.2. 2018
    # Fri Feb 02 2018 00:00:00 == 1517529600 EPOCH
    SHV_EPOCH_SEC = 1517529600
    # ChainPack.INVALID_MIN_OFFSET_FROM_UTC = (-64 * 15)

    @classmethod
    def unpack(cls, data: bytes | str) -> SHVType:
        """Unpack single value from given data."""
        return ChainPackReader.unpack(data)

    @classmethod
    def pack(cls, value: SHVType) -> bytes:
        """Pack given value and return bytes."""
        return ChainPackWriter.pack(value)


class ChainPackReader(commonpack.CommonReader):
    """Read data in ChainPack format."""

    def read_meta(self) -> SHVMetaType | None:
        if self._peek_byte() != ChainPack.CP_MetaMap:
            return None
        self._read_byte()
        return self._read_map()

    def read(self) -> SHVType:
        meta = self.read_meta()

        value: SHVType
        packing_schema = self._read_byte()
        if packing_schema < 128:
            # tiny Int or UInt
            value = packing_schema & 63
            if not packing_schema & 64:
                value = SHVUInt(value)
        elif packing_schema == ChainPack.CP_Null:
            value = None
        elif packing_schema == ChainPack.CP_TRUE:
            value = True
        elif packing_schema == ChainPack.CP_FALSE:
            value = False
        elif packing_schema == ChainPack.CP_Int:
            value = self._read_int_data()
        elif packing_schema == ChainPack.CP_UInt:
            value = SHVUInt(self.read_uint_data())
        elif packing_schema == ChainPack.CP_Double:
            value = self._read_double()
        elif packing_schema == ChainPack.CP_Decimal:
            value = self._read_decimal()
        elif packing_schema == ChainPack.CP_DateTime:
            value = self._read_datetime()
        elif packing_schema == ChainPack.CP_Map:
            value = self._read_map()
        elif packing_schema == ChainPack.CP_IMap:
            value = self._read_map()
        elif packing_schema == ChainPack.CP_List:
            value = self._read_list()
        elif packing_schema == ChainPack.CP_Blob:
            value = self._read_blob()
        elif packing_schema == ChainPack.CP_String:
            value = self._read_string()
        elif packing_schema == ChainPack.CP_CString:
            value = self._read_cstring()
        else:
            raise TypeError(f"ChainPack - Invalid type: {packing_schema}")
        if meta is not None:
            value = SHVMeta.new(value, meta)
        return value

    def _read_uint_dataHelper(self):
        num = 0
        bitlen = 0
        head = self._read_byte()
        if (head & 128) == 0:
            bytes_to_read_cnt = 0
            num = head & 127
            bitlen = 7
        elif (head & 64) == 0:
            bytes_to_read_cnt = 1
            num = head & 63
            bitlen = 6 + 8
        elif (head & 32) == 0:
            bytes_to_read_cnt = 2
            num = head & 31
            bitlen = 5 + 2 * 8
        elif (head & 16) == 0:
            bytes_to_read_cnt = 3
            num = head & 15
            bitlen = 4 + 3 * 8
        else:
            bytes_to_read_cnt = (head & 0xF) + 4
            bitlen = bytes_to_read_cnt * 8

        for _ in range(bytes_to_read_cnt):
            r = self._read_byte()
            num = (num << 8) + r
        return num, bitlen

    def read_uint_data(self):
        num, _ = self._read_uint_dataHelper()
        return num

    def _read_int_data(self):
        num, bitlen = self._read_uint_dataHelper()
        sign_bit_mask = 1 << (bitlen - 1)
        neg = num & sign_bit_mask
        snum = num
        if neg:
            snum &= ~sign_bit_mask
            snum = -snum
        return snum

    def _read_double(self) -> float:
        res = struct.unpack("<d", self._read(8))  # little endian
        return res[0]

    def _read_decimal(self):
        mant = self._read_int_data()
        exp = self._read_int_data()
        return decimal.Decimal(f"{mant}e{exp}")

    def _read_datetime(self):
        d = self.read_uint_data()
        offset = 0
        has_tz_offset = d & 1
        has_not_msec = d & 2
        d >>= 2
        if has_tz_offset:
            offset = d & 0x7F
            if offset >= 128:
                offset -= 128  # sign extension
            d >>= 7
        if not has_not_msec:
            d /= 1000
        d += ChainPack.SHV_EPOCH_SEC
        tzone = datetime.timezone(datetime.timedelta(minutes=offset * 15))
        return datetime.datetime.fromtimestamp(d, tzone)

    def _read_blob(self) -> bytes:
        dlen = self.read_uint_data()
        return self._read(dlen)

    def _read_string(self) -> str:
        slen = self.read_uint_data()
        if slen <= 0:
            return ""
        return self._read(slen).decode("utf-8")

    def _read_cstring(self) -> str:
        res = ""
        while True:
            b = self._read_byte()
            if b == ord("\\"):
                b = self._read_byte()
                if b == ord("\\"):
                    res += "\\"
                elif b == ord("0"):
                    res += "\0"
                else:
                    res += chr(b)
            else:
                if b == 0:
                    break  # end of string
                res += chr(b)
        return res

    def _read_list(self):
        lst = []
        while True:
            b = self._peek_byte()
            if b == ChainPack.CP_TERM:
                self._read_byte()
                break
            lst.append(self.read())
        return lst

    def _read_map(self):
        mmap = {}
        while True:
            b = self._peek_byte()
            if b == ChainPack.CP_TERM:
                self._read_byte()
                break
            key = self.read()
            val = self.read()
            mmap[key] = val
        return mmap


class ChainPackWriter(commonpack.CommonWriter):
    """Write data in ChainPack format."""

    # Integer encoding:
    #  0 ...  7 bits  1  byte  |0|s|x|x|x|x|x|x|<-- LSB
    #  8 ... 14 bits  2  bytes |1|0|s|x|x|x|x|x| |x|x|x|x|x|x|x|x|<-- LSB
    # 15 ... 21 bits  3  bytes |1|1|0|s|x|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x|<-- LSB
    # 22 ... 28 bits  4  bytes |1|1|1|0|s|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x|<-- LSB
    # 29+       bits  5+ bytes |1|1|1|1|n|n|n|n| |s|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x| ... <-- LSB
    #                                         n ==  0 ->  4 bytes number (32 bit number)
    #                                         n ==  1 ->  5 bytes number
    #                                         n == 14 -> 18 bytes number
    #                                         n == 15 -> for future (number of bytes will be specified in next byte)

    @staticmethod
    def _significant_bits_part_length(num: int) -> int:
        length = 0
        if num & 0xFFFFFFFF00000000:
            length += 32
            num >>= 32
        if num & 0xFFFF0000:
            length += 16
            num >>= 16
        if num & 0xFF00:
            length += 8
            num >>= 8
        if num & 0xF0:
            length += 4
            num >>= 4
        # see https://en.wikipedia.org/wiki/Find_first_set#CLZ
        sig_table_4bit = [0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4]
        length += sig_table_4bit[num]
        return max(length, 1)

    @classmethod
    def _bytes_needed(cls, bit_len: int) -> int:
        """Calculate needed maximum number of bytes.

        Return max bit length >= bit_len, which can be encoded by same number of
        bytes number of bytes needed to encode bit_len.
        """
        if bit_len <= 28:
            cnt = ((bit_len - 1) // 7) + 1
        else:
            cnt = ((bit_len - 1) // 8) + 2
        return cnt

    @classmethod
    def _expand_bit_len(cls, bit_len):
        byte_cnt = cls._bytes_needed(bit_len)
        if bit_len <= 28:
            ret = byte_cnt * (8 - 1) - 1
        else:
            ret = (byte_cnt - 1) * 8 - 1
        return ret

    def _write_uint_data_helper(self, num, bit_len):
        byte_cnt = self._bytes_needed(bit_len)
        data = bytearray(byte_cnt)
        for i in range(byte_cnt - 1, -1, -1):
            r = num & 255
            data[i] = r
            num = num >> 8

        if bit_len <= 28:
            mask = 0xF0 << (4 - byte_cnt)
            data[0] = data[0] & ~mask
            mask = (mask << 1) & 0xFF
            data[0] = data[0] | mask
        else:
            data[0] = 0xF0 | (byte_cnt - 5)

        for i in range(0, byte_cnt):
            r = data[i]
            self._write(r)

    def write_meta(self, meta: dict) -> None:
        self._write(ChainPack.CP_MetaMap)
        self._write_map_data(meta)

    def write_null(self) -> None:
        self._write(ChainPack.CP_Null)

    def write_bool(self, value: bool) -> None:
        self._write(ChainPack.CP_TRUE if value else ChainPack.CP_FALSE)

    def write_blob(self, value: bytes | bytearray) -> None:
        self._write(ChainPack.CP_Blob)
        self.write_uint_data(len(value))
        self._write(value)

    def write_string(self, value: str) -> None:
        bstring = value.encode("utf-8")
        self._write(ChainPack.CP_String)
        self.write_uint_data(len(bstring))
        self._write(bstring)

    def write_cstring(self, value: str) -> None:
        bstring = value.encode("utf-8")
        self._write(ChainPack.CP_CString)
        self._write(bstring)
        self._write(b"\0")

    def write_uint(self, value: int) -> None:
        if value < 64:
            self._write(value % 64)
        else:
            self._write(ChainPack.CP_UInt)
            self.write_uint_data(value)

    def write_uint_data(self, value: int) -> None:
        bitcnt = self._significant_bits_part_length(value)
        self._write_uint_data_helper(value, bitcnt)

    def write_int(self, value: int) -> None:
        if 0 <= value < 64:
            self._write((value % 64) + 64)
        else:
            self._write(ChainPack.CP_Int)
            self.write_int_data(value)

    def write_int_data(self, value: int) -> None:
        num = -value if value < 0 else value
        neg = value < 0

        bitlen = self._significant_bits_part_length(num)
        bitlen += 1  # add sign bit
        if neg:
            sign_pos = self._expand_bit_len(bitlen)
            sign_bit_mask = 1 << sign_pos
            num |= sign_bit_mask
        self._write_uint_data_helper(num, bitlen)

    def write_double(self, value: float) -> None:
        self._write(ChainPack.CP_Double)
        self._write(struct.pack("<d", value))  # little endian

    def write_decimal(self, value: decimal.Decimal) -> None:
        n, d = value.as_integer_ratio()
        self._write(ChainPack.CP_Decimal)
        self.write_int_data(n)
        self.write_int_data(-int(math.log10(d)))

    def write_list(self, value: collections.abc.Iterable[SHVType]) -> None:
        self._write(ChainPack.CP_List)
        for val in value:
            self.write(val)
        self._write(ChainPack.CP_TERM)

    def _write_map_data(self, mmap):
        for k, v in mmap.items():
            self.write(k)
            self.write(v)
        self._write(ChainPack.CP_TERM)

    def write_map(self, value: collections.abc.Mapping[str, SHVType]) -> None:
        self._write(ChainPack.CP_Map)
        self._write_map_data(value)

    def write_imap(self, value: collections.abc.Mapping[int, SHVType]) -> None:
        self._write(ChainPack.CP_IMap)
        self._write_map_data(value)

    def write_datetime(self, value: datetime.datetime) -> None:
        self._write(ChainPack.CP_DateTime)
        res = int(value.timestamp() * 1000) - (ChainPack.SHV_EPOCH_SEC * 1000)
        tzdelta = value.utcoffset()
        tzoff = int(tzdelta.total_seconds() // 60 // 15) if tzdelta is not None else 0
        if not -63 <= tzoff <= 63:
            raise TypeError(f"Invalid UTC offset value: {tzoff}")
        ms = res % 1000 == 0
        if ms:
            res //= 1000
        if tzoff != 0:
            res <<= 7
            res |= tzoff & 0x7F
        res <<= 2
        if tzoff != 0:
            res |= 1
        if ms:
            res |= 2
        self.write_int_data(res)



"""RPC client manager that provides facitility to simply connect to the broker."""
import asyncio
import collections.abc
import hashlib
import logging
import time
import traceback
import typing

from .rpcclient import RpcClient
from .rpcerrors import (
    RpcError,
    RpcInvalidParamsError,
    RpcMethodCallExceptionError,
    RpcMethodNotFoundError,
)
from .rpcmessage import RpcMessage
from .rpcmethod import (
    RpcMethodAccess,
    RpcMethodDesc,
    RpcMethodFlags,
    RpcMethodSignature,
)
from .rpcurl import RpcLoginType, RpcUrl
from .value import SHVType, is_shvbool, is_shvnull

logger = logging.getLogger(__name__)


class SimpleClient:
    """SHV client made simple to use.

    You most likely want to use this client instead of using RpcClient directly.

    Messages are handled in an asyncio loop and based on the type of the message
    the different operation is performed.
    """

    IDLE_TIMEOUT: float = 180
    """Number of seconds before we are disconnected from the broker automatically."""

    APP_NAME: str = "pyshv"
    """Name of the application reported to the SHV.

    You can change this value in child class to report a more accurate
    application name.
    """
    APP_VERSION: str = "unknown"
    """Version of the application reported to the SHV.

    You should change this value in child class to report a correct number.
    """

    def __init__(self, client: RpcClient, client_id: int | None):
        self.client = client
        """The underlaying RPC client instance."""
        self.client_id = client_id
        """Client ID assigned on login by SHV broker."""
        self.task = asyncio.create_task(self._loop())
        """Task running the message handling loop."""
        self._calls_event: dict[int, asyncio.Event] = {}
        self._calls_msg: dict[int, RpcMessage] = {}

    @classmethod
    async def connect(
        cls,
        url: RpcUrl,
        login_options: dict[str, SHVType] | None = None,
    ) -> typing.Union["SimpleClient", None]:
        """Connect and login to the SHV broker.

        :param url: SHV RPC URL to the broker
        :param login_options: Additional options sent with login
        """
        client = await RpcClient.connect(url.host, url.port, url.protocol)
        options = url.login_options()
        if login_options:
            options.update(login_options)
        cid = await cls.login(
            client, url.username, url.password, url.login_type, options
        )
        if client.writer.is_closing():
            return None
        return cls(client, cid)

    async def disconnect(self) -> None:
        """Disconnect an existing connection.

        The call to the disconnect when client is not connected is silently
        ignored.
        """
        if self.client.writer.is_closing():
            return
        await self.client.disconnect()
        await self.task

    @classmethod
    async def login(
        cls,
        client: RpcClient,
        username: str | None,
        password: str | None,
        login_type: RpcLoginType,
        login_options: dict[str, SHVType] | None,
    ) -> int | None:
        """Perform login to the broker.

        The login need to be performed only once right after the connection is
        established.

        :return: Client ID assigned by broker or `None` in case none was
            assigned.
        """
        # Note: The implementation here expects that broker won't sent any other
        # messages until login is actually performed. That is what happens but
        # it is not defined in any SHV design document as it seems.
        await client.send(RpcMessage.request(None, "hello"))
        resp = await client.receive()
        if resp is None:
            return None
        resl = resp.result()
        assert isinstance(resl, collections.abc.Mapping)

        nonce = resl.get("nonce", None)
        if login_type is RpcLoginType.SHA1:
            assert isinstance(nonce, str)
            m = hashlib.sha1()
            m.update(nonce.encode("utf-8"))
            m.update((password or "").encode("utf-8"))
            password = m.hexdigest()
        params: SHVType = {
            "login": {"password": password, "type": login_type.value, "user": username},
            "options": {
                "idleWatchDogTimeOut": int(cls.IDLE_TIMEOUT),
                **(login_options if login_options else {}),  # type: ignore
            },
        }

        logger.debug("LOGGING IN")
        await client.send(RpcMessage.request(None, "login", params))
        resp = await client.receive()
        if resp is None:
            return None
        result = resp.result()
        logger.debug("LOGGED IN")
        cid = result.get("clientId", None) if isinstance(result, dict) else None
        if isinstance(cid, int):
            return cid
        return 0

    async def _loop(self) -> None:
        """Loop run in asyncio task to receive messages."""
        keep_alive_task = asyncio.create_task(self._keep_alive_loop())
        while msg := await self.client.receive(raise_error=False):
            asyncio.create_task(self._message(msg))
        keep_alive_task.cancel()
        try:
            await keep_alive_task
        except asyncio.exceptions.CancelledError:
            pass

    async def _keep_alive_loop(self) -> None:
        """Loop run alongside with _loop to send pings to the broker when idling."""
        idlet = self.IDLE_TIMEOUT / 2
        while not self.client.writer.is_closing():
            t = time.monotonic() - self.client.last_activity
            if t < idlet:
                await asyncio.sleep(idlet - t)
            else:
                await self.client.send(RpcMessage.request(".broker/app", "ping"))

    async def _message(self, msg: RpcMessage) -> None:
        """Handle every received message."""
        if msg.is_request():
            resp = msg.make_response()
            method = msg.method()
            assert method  # is ensured by is_request but not detected by mypy
            try:
                resp.set_result(
                    await self._method_call(
                        msg.shv_path() or "",
                        method,
                        msg.rpc_access_grant() or RpcMethodAccess.BROWSE,
                        msg.params(),
                    )
                )
            except RpcError as exp:
                resp.set_rpc_error(exp)
            except Exception as exc:
                resp.set_rpc_error(
                    RpcMethodCallExceptionError(
                        "".join(traceback.format_exception(exc))
                    )
                )
            await self.client.send(resp)
        elif msg.is_response():
            rid = msg.request_id()
            assert rid is not None
            if rid in self._calls_event:
                self._calls_msg[rid] = msg
                self._calls_event.pop(rid).set()
        elif msg.is_signal():
            if msg.method() == "chng":
                await self._value_update(msg.shv_path() or "", msg.params())

    async def call(self, path: str, method: str, params: SHVType = None) -> SHVType:
        """Call given method on given path with given parameters.

        Note that this is coroutine and this it is up to you if you await it or
        use asyncio tasks.

        :param path: SHV path method is associated with.
        :param method: SHV method name to be called.
        :param params: Parameters passed to the called method.
        :return: Return value on successful method call.
        :raise RpcError: The call result in error that is propagated by raising
            `RpcError` or its children based on the failure.
        """
        msg = RpcMessage.request(path, method, params)
        rid = msg.request_id()
        assert rid is not None
        event = asyncio.Event()
        self._calls_event[rid] = event
        await self.client.send(msg)
        await event.wait()
        msg = self._calls_msg.pop(rid)
        err = msg.rpc_error()
        if err is not None:
            raise err
        return msg.result()

    async def ls(self, path: str) -> list[str]:
        """List child nodes of the node on the specified path.

        :param path: SHV path to the node we want children to be listed for.
        :return: list of child nodes.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "ls")
        if not isinstance(res, list):
            raise RpcMethodCallExceptionError(f"Invalid result returned: {repr(res)}")
        return res

    async def ls_with_children(self, path: str) -> dict[str, bool]:
        """List child nodes of the node on the specified path.

        Compared to the :func:`ls` method this provides you also with info on
        children of the node. This can safe unnecessary listing requests when
        you are iterating over the tree.

        :param path: SHV path to the node we want children to be listed for.
        :return: dictionary where keys are names of the nodes and values are
            booleans signaling presence of at least one child.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "ls", ("", 1))
        if not isinstance(res, list):
            raise RpcMethodCallExceptionError(f"Invalid result returned: {repr(res)}")
        return {v[0]: v[1] for v in res}

    async def dir(self, path: str) -> list[str]:
        """List methods associated with node on the specified path.

        :param path: SHV path to the node we want methods to be listed for.
        :return: list of the node's methods.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "dir", ("", 0))
        if not isinstance(res, list):
            raise RpcMethodCallExceptionError(f"Invalid result returned: {repr(res)}")
        return [m for m in res if isinstance(m, str)]

    async def dir_details(self, path: str) -> list[RpcMethodDesc]:
        """List methods associated with node on the specified path.

        :param path: SHV path to the node we want methods to be listed for.
        :return: list of the node's method descriptors.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "dir")
        if not isinstance(res, list):
            raise RpcMethodCallExceptionError(f"Invalid result returned: {repr(res)}")
        return [RpcMethodDesc.frommap(m) for m in res]

    async def subscribe(self, path: str) -> None:
        """Perform subscribe for signals on given path.

        Subscribe is always performed on the node itself as well as all its
        children.

        :param path: SHV path to the node to subscribe.
        """
        await self.call(".broker/app", "subscribe", {"path": path})

    async def unsubscribe(self, path: str) -> bool:
        """Perform unsubscribe for signals on given path.

        :param path: SHV path previously passed to :func:`subscribe`.
        :return: ``True`` in case such subscribe was located and ``False``
            otherwise.
        """
        resp = await self.call(".broker/app", "unsubscribe", {"path": path})
        assert is_shvbool(resp)
        return bool(resp)

    async def _method_call(
        self, path: str, method: str, access: RpcMethodAccess, params: SHVType
    ) -> SHVType:
        """Handle request in the provided message.

        You have to set some RpcMessage that is sent back as a response.

        :param path: SHV path to the node the method is associated with.
        :param method: method requested to be called.
        :param access: access level of the client specified in the request.
        :param params: Parameters to be passed to the called method.
        :return: result of the method call. To report error you should raise
            :exc:`RpcError`.
        """
        if method == "ls":
            return self._method_call_ls(path, params)
        if method == "dir":
            return self._method_call_dir(path, params)
        if method == "appName":
            return self.APP_NAME
        if method == "appVersion":
            return self.APP_VERSION
        if method == "echo" and access >= RpcMethodAccess.WRITE:
            return params
        raise RpcMethodNotFoundError(
            f"No such path '{path}' or method '{method}' or access rights."
        )

    def _method_call_ls(self, path: str, params: SHVType) -> SHVType:
        """Implementation of the ``ls`` method call."""
        if not is_shvnull(params) and not (
            isinstance(params, list)
            and len(params) == 2
            and isinstance(params[0], str)
            and isinstance(params[1], int)
        ):
            raise RpcInvalidParamsError("Use Null or list with path and attributes")
        children = False
        if isinstance(params, list):
            path = "/".join(filter(lambda v: v, (path, params[0])))
            if params[1] & 0x1:
                children = True
        lsres = self._ls(path)
        if not children:
            return list(val[0] for val in lsres)
        return list(lsres)

    def _ls(self, path: str) -> typing.Iterator[tuple[str, bool | None]]:
        """Implement ``ls`` method for all nodes.

        The default implementation returns empty list for the root and raises
        exception :exc:`RpcMethodNotFoundError` otherwise. Your implementation
        should return list of child nodes. Every node should be tuple with its
        name and boolean signaling if there are some children.

        You should call this method as a last resort as it will most likely
        raise error.

        :param path: SHV path that should be listed.
        :return: Iterator over child nodes of the node on the path.
        :raise RpcMethodNotFoundError: when there is no such method. This has to
            be on first iteration and thus you should not yield any value if you
            plan to raise this exception!
        """
        if path:
            raise RpcMethodNotFoundError(f"No such node: {path}")
        return iter(())

    def _method_call_dir(self, path: str, params: SHVType) -> SHVType:
        if (
            not is_shvnull(params)
            and not isinstance(params, str)
            and not (
                isinstance(params, list)
                and len(params) == 2
                and isinstance(params[0], str)
                and isinstance(params[1], int)
            )
        ):
            raise RpcInvalidParamsError(
                "Use Null, string or list with path and attributes"
            )
        path = "/".join(
            filter(
                lambda v: v,
                (
                    path,
                    params if isinstance(params, str) else "",
                    params[0] if isinstance(params, list) else "",
                ),
            )
        )
        try:
            next(self._ls(path))  # detect missing node (raises RpcMethodNotFoundError)
        except StopIteration:
            pass
        dirres = self._dir(path)
        if not isinstance(params, list) or params[1] == 127:
            return [d.tomap() for d in dirres]
        return list(d.name for d in dirres)

    def _dir(self, path: str) -> typing.Iterator[RpcMethodDesc]:
        """Implement ``dir`` method for all nodes.

        This implementation is called only when :meth:`_ls` does not raise
        exception and thus only for valid paths.

        Always call this as first before you yield your own methods to provide
        users with standard ones.

        :param path: SHV path method should be listed for.
        :return: List of methods associated with given node.
        """
        yield RpcMethodDesc("dir", RpcMethodSignature.RET_PARAM)
        yield RpcMethodDesc("ls", RpcMethodSignature.RET_PARAM)
        if not path:
            yield RpcMethodDesc(
                "appName", RpcMethodSignature.RET_VOID, RpcMethodFlags.GETTER
            )
            yield RpcMethodDesc(
                "appVersion", RpcMethodSignature.RET_VOID, RpcMethodFlags.GETTER
            )
            yield RpcMethodDesc(
                "echo",
                RpcMethodSignature.RET_PARAM,
                access=RpcMethodAccess.WRITE,
            )

    async def _value_update(self, path: str, value: SHVType) -> None:
        """Handle value change (`chng` method)."""



"""Perform tests of our examples."""
import pytest

import example_client
from shv import RpcMethodNotFoundError, shvmeta_eq

LS_DATA: list[tuple[str, list[tuple[str, bool | None]]]] = [
    ("", [(".broker", True), ("test", True)]),
    ("test", [("device", None)]),
    ("test/device", [("track", True)]),
    ("test/device/track", [(str(i), False) for i in range(1, 9)]),
]


@pytest.mark.parametrize(
    "path,result", [(c[0], list(v[0] for v in c[1])) for c in LS_DATA]
)
async def test_ls(example_device, client, path, result):
    res = await client.call(path, "ls")
    assert shvmeta_eq(res, result)


@pytest.mark.parametrize("path,result", LS_DATA)
async def test_ls_with_children(example_device, client, path, result):
    res = await client.call(path, "ls", ("", 1))
    assert shvmeta_eq(res, result)


DIR_DATA = [
    (
        "test/device",
        [
            {"accessGrant": "bws", "flags": 0, "signature": 3, "name": "dir"},
            {"accessGrant": "bws", "flags": 0, "signature": 3, "name": "ls"},
            {"accessGrant": "bws", "flags": 2, "name": "appName", "signature": 2},
            {
                "accessGrant": "bws",
                "flags": 2,
                "name": "appVersion",
                "signature": 2,
            },
            {"accessGrant": "wr", "flags": 0, "name": "echo", "signature": 3},
        ],
    ),
    (
        "test/device/track",
        [
            {"accessGrant": "bws", "flags": 0, "signature": 3, "name": "dir"},
            {"accessGrant": "bws", "flags": 0, "signature": 3, "name": "ls"},
        ],
    ),
    (
        "test/device/track/1",
        [
            {"accessGrant": "bws", "flags": 0, "signature": 3, "name": "dir"},
            {"accessGrant": "bws", "flags": 0, "signature": 3, "name": "ls"},
            {
                "accessGrant": "rd",
                "flags": 2,
                "signature": 2,
                "name": "get",
                "description": "Get current track",
            },
            {
                "accessGrant": "wr",
                "flags": 4,
                "signature": 1,
                "name": "set",
                "description": "Set track",
            },
        ],
    ),
]


@pytest.mark.parametrize(
    "path,result", [(c[0], list(v["name"] for v in c[1])) for c in DIR_DATA]
)
async def test_dir(example_device, client, path, result):
    res = await client.call(path, "dir", ("", 0))
    assert shvmeta_eq(res, result)


@pytest.mark.parametrize("path,result", DIR_DATA)
async def test_dir_details(example_device, client, path, result):
    res = await client.call(path, "dir")
    assert shvmeta_eq(res, result)


async def test_get(example_device, client):
    res = await client.call("test/device/track/4", "get")
    assert res == [0, 1, 2, 3]


async def test_set(example_device, client):
    tracks = [3, 2, 1, 0]
    res = await client.call("test/device/track/4", "set", tracks)
    assert res is True

    res = await client.call("test/device/track/4", "get")
    assert shvmeta_eq(res, tracks)


async def test_invalid_request(example_device, client):
    with pytest.raises(RpcMethodNotFoundError):
        await client.call("test/device/track/4", "nosuchmethod")


async def test_no_such_node(example_device, client):
    with pytest.raises(RpcMethodNotFoundError):
        await client.ls("test/device/track/none")
    with pytest.raises(RpcMethodNotFoundError):
        await client.dir("test/device/track/none")


async def test_example_client(shvbroker, url):
    await example_client.example_client(url)



"""Common extension for the SimpleClient."""
import collections.abc
import datetime
import typing

from .rpcclient import RpcClient
from .simpleclient import SimpleClient
from .value import SHVType, shvmeta, shvmeta_eq


class ValueClient(SimpleClient, collections.abc.Mapping):
    """SHV client made to track values more easily.

    This tailors to the use case of tracking and accessing various values more
    easily. You need to subscribe to specific path and this class automatically
    provides you with cached latest value as received through signals or fetched
    from logs (logs fetching has to be performed explicitly).

    To access subscribed value you can index this object with SHV path to it.
    """

    def __init__(self, client: RpcClient, client_id: int | None):
        super().__init__(client, client_id)
        self._cache: dict[str, SHVType] = {}
        self._handlers: dict[str, typing.Callable] = {}

    def __getitem__(self, key: str) -> SHVType:
        return self._cache[key]

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self._cache.keys())

    def __len__(self):
        return len(self._cache)

    async def _value_update(self, path: str, value: SHVType) -> None:
        handler = self._get_handler(path)
        if handler is not None and not shvmeta_eq(self._cache.get(path, None), value):
            handler(self, path, value)
        self._cache[path] = value

    def _get_handler(self, path: str) -> None | typing.Callable:
        """Get the handler for the longest path match."""
        split_key = path.split("/")
        paths = (
            "/".join(split_key[: len(split_key) - i]) for i in range(len(split_key))
        )
        return next(
            (self._handlers[path] for path in paths if path in self._handlers),
            None,
        )

    async def prop_get(self, path: str) -> SHVType:
        """Get value from the property associated with the node on given path."""
        return await self.call(path, "get")

    async def prop_set(self, path: str, value: SHVType) -> None:
        """Set value to the property associated with the node on given path."""
        await self.call(path, "set", value)

    def on_change(self, path: str, callback: typing.Callable | None) -> None:
        """Register new callback handler called when value changes.

        The handler is called right before value is updated and thus it is
        possible to access the old and new value is provided as an argument.

        To clear the callback on the specific path you can pass `None`.

        :param path:
        :param callback:
        """
        if callback is None:
            self._handlers.pop(path)
        else:
            self._handlers[path] = callback

    async def log_snapshot(self, path: str) -> None:
        """Get snapshot of the logs.

        Use this to receive old values.

        :param path: SHV path to the node with `getLog` method.
        """
        params: SHVType = {
            "recordCountLimit": 10000,
            "withPathsDict": True,
            "withSnapshot": True,
            "withTypeInfo": False,
            "since": datetime.datetime.now(),
        }
        result = await self.call(path, "getLog", params)
        if result:
            paths_dict = shvmeta(result).get("pathsDict", None)
            if isinstance(paths_dict, collections.abc.Sequence):
                for list_item in paths_dict:
                    if not isinstance(list_item, collections.abc.Sequence):
                        continue
                    idx = list_item[1]
                    if not isinstance(idx, int):
                        continue
                    value = list_item[2]
                    spath = paths_dict[idx]
                    if not isinstance(spath, str):
                        continue
                    await self._value_update(spath, value)

    async def get_snapshot(self, path: str) -> None:
        """Get snapshot of data using get methods.

        This provides a way for you to initialize cache without logs. It
        iterates over SHV tree and calls any get method it encounters.
        """
        # TODO rewrite this to use ls_with_children to safe on method calls
        if "get" in await self.dir(path) and path not in self._cache:
            self._cache[path] = await self.prop_get(path)
        for node in await self.ls(path):
            await self.get_snapshot("/".join((path, node)))



"""Implementation of RpcMessage."""
from __future__ import annotations

import collections.abc
import enum
import typing

from .chainpack import ChainPackWriter
from .cpon import CponWriter
from .rpcerrors import RpcError, RpcErrorCode
from .rpcmethod import RpcMethodAccess
from .value import SHVDict, SHVMeta, SHVType, is_shvimap, is_shvmap, shvmeta


class RpcMessage:
    """Single SHV RPC message representation."""

    last_request_id: typing.ClassVar[int] = 0
    """Counter of request IDs to ensure that every request has unique ID."""

    @classmethod
    def next_request_id(cls) -> int:
        """Provides you with unique request identifier.

        The identifier won't repeat for this application as it is just simple
        counter that should never wrap.
        """
        cls.last_request_id += 1
        return cls.last_request_id

    def __init__(self, rpc_val=None):
        self.value = SHVMeta.new({}, {}) if rpc_val is None else rpc_val

    class Tag(enum.IntEnum):
        """Tags in Meta for RPC message."""

        REQUEST_ID = 8
        SHVPATH = 9
        METHOD = 10
        CALLER_IDS = 11
        ACCESS_GRANT = 14

    class Key(enum.IntEnum):
        """Keys in the toplevel IMap of the RPC message."""

        PARAMS = 1
        RESULT = 2
        ERROR = 3

    class ErrorKey(enum.IntEnum):
        """Keys in the error IMap."""

        CODE = 1
        MESSAGE = 2

    @classmethod
    def request(
        cls,
        path: str | None,
        method: str,
        params: SHVType = None,
        rid: int | None = None,
    ) -> "RpcMessage":
        """Create request message.

        :param path: SHV path for signal.
        :param method: method name for signal.
        :param params: Parameters passed to the method.
        :param rid: Request identifier for this message. It is automatically assigned if
          ``None`` is passed.
        """
        res = cls()
        res.set_request_id(rid or cls.next_request_id())
        res.set_method(method)
        res.set_shv_path(path)
        res.set_params(params)
        return res

    @classmethod
    def signal(cls, path, method, value) -> "RpcMessage":
        """Create signal message.

        :param path: SHV path for signal.
        :param method: method name for signal.
        :param value: Value to be sent in the message.
        """
        res = cls()
        res.set_method(method)
        res.set_shv_path(path)
        res.set_params(value)
        return res

    @classmethod
    def chng(cls, path, value) -> "RpcMessage":
        """Create message for ``chng`` signal.

        :param path: SHV path for signal.
        :param value: New value to be sent in the message.
        """
        return cls.signal(path, "chng", value)

    def is_valid(self) -> bool:
        """Check if message is valid RPC message."""
        # TODO maybe do more work than just check basic type
        return isinstance(self.value, SHVDict)

    def is_request(self):
        """Check if message is request."""
        return self.request_id() and self.method()

    def is_response(self):
        """Check if message is a response."""
        return self.request_id() and not self.method()

    def is_error(self):
        """Check if message is an error response."""
        return (
            self.is_response()
            and self.error() is not None
            and self.shverror()[0] != RpcErrorCode.NO_ERROR
        )

    def is_signal(self):
        """Check if message is a signal."""
        return not self.request_id() and self.method()

    def make_response(self) -> RpcMessage:
        if not self.is_request():
            raise ValueError("Response can be created from request only.")
        resp = RpcMessage()
        resp.set_request_id(self.request_id())
        resp.set_caller_ids(self.caller_ids())
        return resp

    def request_id(self) -> int | None:
        """Request identificator of this message."""
        res = shvmeta(self.value).get(self.Tag.REQUEST_ID, None)
        return res if isinstance(res, int) else None

    def set_request_id(self, rqid: int | None) -> None:
        """Set given request idenficator to this message."""
        if rqid is None:
            shvmeta(self.value).pop(self.Tag.REQUEST_ID, None)
        else:
            self.value = SHVMeta.new(self.value, {self.Tag.REQUEST_ID: rqid})

    def shv_path(self) -> str | None:
        """SHV path specified for this message."""
        res = shvmeta(self.value).get(self.Tag.SHVPATH, None)
        return res if isinstance(res, str) else None

    def set_shv_path(self, path: str | None) -> None:
        """Set given path as SHV path for this message."""
        if path is None:
            shvmeta(self.value).pop(self.Tag.SHVPATH, None)
        else:
            self.value = SHVMeta.new(self.value, {self.Tag.SHVPATH: path})

    def caller_ids(self) -> collections.abc.Sequence[int] | None:
        """Caller idenfieiers associated with this message."""
        res = shvmeta(self.value).get(self.Tag.CALLER_IDS, None)
        if isinstance(res, int):
            return [res]
        if not isinstance(res, list):
            return None
        filter(lambda v: isinstance(v, int), res)
        return res  # type: ignore

    def set_caller_ids(self, cids: collections.abc.Sequence[int] | None) -> None:
        """Set caller idenfieiers associated with this message."""
        if not cids:
            shvmeta(self.value).pop(self.Tag.CALLER_IDS, None)
        else:
            some: SHVType = cids
            if len(cids) == 1:
                some = cids[0]
            self.value = SHVMeta.new(self.value, {self.Tag.CALLER_IDS: some})

    def access_grant(self) -> list[str] | None:
        """Granted access level of the caller."""
        res = shvmeta(self.value).get(self.Tag.ACCESS_GRANT, None)
        if isinstance(res, str):
            return res.split(",")
        return None

    def set_access_grant(self, access: collections.abc.Sequence[str]) -> None:
        """Granted access level of the caller."""
        if not access:
            shvmeta(self.value).pop(self.Tag.ACCESS_GRANT, None)
        else:
            self.value = SHVMeta.new(
                self.value, {self.Tag.ACCESS_GRANT: ",".join(access)}
            )

    def rpc_access_grant(self) -> RpcMethodAccess | None:
        """Granted access level of the caller as :class:`shv.RpcMethodAccess`."""
        m = RpcMethodAccess.strmap()
        for access in self.access_grant() or []:
            if access in m:
                return m[access]
        return None

    def set_rpc_access_grant(self, access: RpcMethodAccess | None) -> None:
        """Set granted access level of the caller."""
        if access is None:
            shvmeta(self.value).pop(self.Tag.ACCESS_GRANT, None)
        else:
            self.value = SHVMeta.new(
                self.value, {self.Tag.ACCESS_GRANT: RpcMethodAccess.tostr(access)}
            )

    def method(self) -> str | None:
        """SHV method name for this message."""
        res = shvmeta(self.value).get(self.Tag.METHOD, None)
        return res if isinstance(res, str) else None

    def set_method(self, method: str) -> None:
        """Set SHV method name for this message."""
        if method is None:
            shvmeta(self.value).pop(self.Tag.METHOD, None)
        else:
            self.value = SHVMeta.new(self.value, {self.Tag.METHOD: method})

    def params(self) -> SHVType:
        """SHV parameters for the method call."""
        return self.value.get(self.Key.PARAMS, None) if is_shvimap(self.value) else None

    def set_params(self, params: SHVType) -> None:
        """Set SHV parameters for this method call."""
        if params is None:
            self.value.pop(self.Key.PARAMS, None)
        else:
            self.value[self.Key.PARAMS] = params

    def result(self) -> SHVType:
        """SHV method call result."""
        return self.value.get(self.Key.RESULT, None) if is_shvimap(self.value) else None

    def set_result(self, result: SHVType) -> None:
        """Set SHV method call result."""
        if result is None:
            self.value.pop(self.Key.RESULT, None)
        else:
            self.value[self.Key.RESULT] = result

    def error(self) -> SHVType:
        """SHV method call error."""
        return self.value.get(self.Key.ERROR, None) if is_shvimap(self.value) else None

    def set_error(self, error: SHVType) -> None:
        """Set SHV method call error."""
        if error is None:
            self.value.pop(self.Key.ERROR, None)
        else:
            self.value[self.Key.ERROR] = error

    def shverror(self) -> tuple[RpcErrorCode, str]:
        """SHV method call error in standard SHV format."""
        res = self.error()
        if is_shvimap(res):
            assert isinstance(res, dict)
            rcode = res.get(self.ErrorKey.CODE)
            code: RpcErrorCode = RpcErrorCode.UNKNOWN
            if isinstance(rcode, int):
                try:
                    code = RpcErrorCode(rcode)
                except ValueError:
                    pass
            rmsg = res.get(self.ErrorKey.MESSAGE)
            msg = rmsg if isinstance(rmsg, str) else ""
            return code, msg
        return RpcErrorCode.UNKNOWN, res if isinstance(res, str) else ""

    def set_shverror(self, code: RpcErrorCode, msg: str = ""):
        """Set SHV method call error in standard SHV format."""
        self.set_error({self.ErrorKey.CODE: code, self.ErrorKey.MESSAGE: msg})

    def rpc_error(self) -> RpcError | None:
        """Get SHV RPC error as RpcError."""
        if not self.is_error():
            return None
        code, msg = self.shverror()
        return RpcError(msg, code)

    def set_rpc_error(self, error: RpcError) -> None:
        """Set given SHV RPC error to this message."""
        self.set_shverror(error.error_code, error.message)

    def to_string(self) -> str:
        """Convert message to CPON and return it as string."""
        return self.to_cpon().decode("utf-8")

    def to_cpon(self) -> bytes:
        """Convert message to Cpon."""
        return CponWriter.pack(self.value) if self.is_valid() else b""

    def to_chainpack(self) -> bytes:
        """Convert message to Chainpack."""
        return ChainPackWriter.pack(self.value) if self.is_valid() else b""



"""SHV RPC URL for RpcClient and RpcServer."""
import dataclasses
import enum
import urllib.parse

from .value import SHVType


class RpcProtocol(enum.Enum):
    """Enum of supported RPC protocols by this Python implementation."""

    TCP = enum.auto()
    LOCAL_SOCKET = enum.auto()


class RpcLoginType(enum.Enum):
    """Enum specifying which login type should be used.

    The string values are the exact string representation used in SHV RPC
    protocol identifying these login types.
    """

    PLAIN = "PLAIN"
    """Plain login format should be used."""
    SHA1 = "SHA1"
    """Use hash algorithm SHA1 (preferred and common default)."""


@dataclasses.dataclass
class RpcUrl:
    """SHV RPC URL specifier.

    This is unified locator for SHV RPC connections that is used to specify
    connections. It is implemented as :func:`dataclasses.dataclass` and you
    can set properties directly or you can parse string URL.
    """

    # URL primary fields
    host: str
    """Hostname of the SHV RPC server."""
    port: int = 3755
    """Port the SHV RPC server is listening on."""
    protocol: RpcProtocol = RpcProtocol.TCP
    """SHV RPC protocol used to communicate (This is scheme in URL therminology)"""
    username: str = ""
    """User name used to login to the remote server."""

    # Options
    password: str | None = None
    """Password used to login to the server."""
    login_type: RpcLoginType = RpcLoginType.SHA1
    """Type of the login to be used (specifies format of the password)."""
    device_id: str | None = None
    """Device identifier sent to the server with login."""
    device_mount_point: str | None = None
    """Request for mounting of connected device to the specified mount point."""

    def login_options(self) -> dict[str, SHVType]:
        """Assemble login options for the SHV RPC broker from options here."""
        res: dict[str, SHVType] = {}
        if self.device_id:
            res["device"] = res.get("device", {})
            res["device"]["deviceId"] = self.device_id  # type: ignore
        if self.device_mount_point:
            res["device"] = res.get("device", {})
            res["device"]["mountPoint"] = self.device_mount_point  # type: ignore
        return res

    @classmethod
    def parse(cls, url: str) -> "RpcUrl":
        """Parse string URL to the object representation.

        :param url: URL in string format.
        :return: New :class:`RpcUrl` instance.
        """
        sr = urllib.parse.urlsplit(url, scheme="localsocket", allow_fragments=False)
        pqs = urllib.parse.parse_qs(sr.query)

        protocols = {
            "tcp": RpcProtocol.TCP,
            "localsocket": RpcProtocol.LOCAL_SOCKET,
        }
        if sr.scheme not in protocols:
            raise ValueError(f"Invalid scheme: {sr.scheme}")
        protocol = protocols[sr.scheme]

        res = cls("", protocol=protocol)

        if protocol is RpcProtocol.TCP:
            res.username = sr.username or ""
            res.host = sr.hostname or ""
            if sr.port is not None:
                res.port = int(sr.port)
        elif protocol is RpcProtocol.LOCAL_SOCKET:
            res.host = f"/{sr.netloc}{sr.path}" if sr.netloc else sr.path
        else:
            raise NotImplementedError()  # pragma: no cover

        if opts := pqs.pop("devid", []):
            res.device_id = opts[0]
        if opts := pqs.pop("devmount", []):
            res.device_mount_point = opts[0]
        if protocol in (RpcProtocol.TCP,):
            # We prefer SHA1 password and thus discard plain if both are present
            if opts := pqs.pop("shapass", []):
                res.password = opts[0]
                res.login_type = RpcLoginType.SHA1
            elif opts := pqs.pop("password", []):
                res.password = opts[0]
                res.login_type = RpcLoginType.PLAIN

        if pqs:
            raise ValueError(f"Unsupported URL queries: {pqs.keys()}")

        return res

    def to_url(self) -> str:
        """Convert to string URL."""
        protocols = {
            RpcProtocol.TCP: "tcp",
            RpcProtocol.LOCAL_SOCKET: "localsocket",
        }
        if self.protocol is RpcProtocol.TCP:
            netloc = "//"
            if self.username:
                netloc += f"{self.username}@"
            if ":" in self.host:
                netloc += f"[{self.host}]"
            else:
                netloc += self.host
            netloc += f":{self.port}"
        elif self.protocol is RpcProtocol.LOCAL_SOCKET:
            netloc = self.host
        else:
            raise NotImplementedError()  # pragma: no cover

        opts: list[str] = []
        if self.device_id:
            opts.append(f"devid={self.device_id}")
        if self.device_mount_point:
            opts.append(f"devmount={self.device_mount_point}")
        if self.protocol in (RpcProtocol.TCP,) and self.password:
            if self.login_type is RpcLoginType.SHA1:
                # TODO escape password string?
                opts.append(f"shapass={self.password}")
            elif self.login_type is RpcLoginType.PLAIN:
                opts.append(f"password={self.password}")
            else:
                raise NotImplementedError()  # pragma: no cover

        return (
            f"{protocols[self.protocol]}:{netloc}{'?' if opts else ''}{'&'.join(opts)}"
        )



"""Common state for the RPC broker that works as server in the SHV network."""
import asyncio
import collections.abc
import dataclasses
import enum
import importlib.metadata
import logging
import random
import time
import typing

from ..rpcclient import RpcClient
from ..rpcerrors import RpcErrorCode
from ..rpcmessage import RpcMessage
from ..rpcmethod import RpcMethodAccess, RpcMethodDesc, RpcMethodSignature
from ..rpcserver import create_rpc_server
from ..rpcurl import RpcLoginType
from ..simpleclient import SimpleClient
from ..value import SHVType, shvget
from .rpcbrokerconfig import RpcBrokerConfig

logger = logging.getLogger(__name__)


class RpcBroker:
    """SHV RPC broker.

    The broker manages multiple RpcClient instances and exchanges messages
    between them.
    """

    def __init__(self, config: RpcBrokerConfig) -> None:
        self.clients: dict[int, RpcBroker.Client] = {}
        self.next_caller_id = 0
        self.config = config
        self.servers: dict[str, asyncio.Server] = {}

    def add_client(self, client: RpcClient):
        """Add a new client to be handled by the broker.

        :param client: RPC client instance to be handled by broker.
        """
        # TODO try to reuse older caller IDs to send smaller messages
        cid = self.next_caller_id
        self.next_caller_id += 1
        self.clients[cid] = self.Client(client, None, self, cid)
        logger.info("Client registered to broker with ID: %s", cid)

    def peer_on_path(self, path: str) -> tuple["RpcBroker.Client", str] | None:
        """Locate client mounted on given path.

        :return: client associated with this mount point and path relative to
            the client or None if there is no such client.
        """
        pth = path.split("/")

        if (
            len(pth) >= 4
            and pth[0] == ".broker"
            and pth[1] == "clients"
            and pth[3] == "app"
        ):
            try:
                client = self.clients[int(pth[2], 10)]
            except (ValueError, KeyError):
                return None
            return (client, "/".join(pth[4:])) if client else None

        # Note: we do not allow recursive mount points and thus first match is the
        # correct and the only client.
        for c in self.clients.values():
            if c.mount_point and pth[: len(c.mount_point)] == c.mount_point:
                return c, "/".join(pth[len(c.mount_point) :])
        return None

    async def start_serving(self) -> None:
        """Start accepting connectons on all configured servers."""
        for name, url in self.config.listen.items():
            if name not in self.servers:
                self.servers[name] = await create_rpc_server(self.add_client, url)

    async def serve_forever(self) -> None:
        """Serve all configured servers."""
        await self.start_serving()
        await asyncio.gather(
            *(server.serve_forever() for server in self.servers.values()),
            return_exceptions=True,
        )
        # TODO handle returned errors

    def close(self) -> None:
        """Request stop listening.

        This stops all servers and thus they no longer accept a new connections but the
        old connections are still kept and working.
        """
        for server in self.servers.values():
            server.close()

    async def wait_closed(self) -> None:
        """Wait for close to complete."""
        await asyncio.gather(
            *(server.wait_closed() for server in self.servers.values()),
            return_exceptions=True,
        )

    async def terminate(self) -> None:
        """Request termination of the broker.

        This closes broker as well as disconnects all established clients.
        """
        self.close()
        await self.wait_closed()
        await asyncio.gather(*(client.disconnect() for client in self.clients.values()))

    class Client(SimpleClient):
        """Single client connected to the broker."""

        APP_NAME = "pyshvbroker"
        try:
            APP_VERSION = importlib.metadata.version("pyshv")
        except importlib.metadata.PackageNotFoundError:
            APP_VERSION = "unknown"

        class State(enum.Enum):
            """State of the client."""

            CONNECTED = enum.auto()
            HELLO = enum.auto()
            LOGIN = enum.auto()

        @dataclasses.dataclass(frozen=True)
        class Subscription:
            """Single subscription."""

            path: str
            method: str

        def __init__(
            self,
            client: RpcClient,
            client_id: int | None,
            broker: "RpcBroker",
            broker_client_id: int,
        ):
            super().__init__(client, client_id)
            self.broker: RpcBroker = broker
            """Broker this client is part of."""
            self.broker_client_id: int = broker_client_id
            """Identifier assigned to the client in the broker."""
            self.mount_point: list[str] = []
            """Mount point for this client split by '/'."""
            self.subscriptions: set[RpcBroker.Client.Subscription] = set()
            """Set of all subscriptions."""
            self.state = self.State.CONNECTED
            """Specifies state of the client login process."""
            self.user: RpcBrokerConfig.User | None = None
            """User used to login to the broker."""
            self._nonce = "".join(random.choice("0123456789abcdef") for _ in range(8))

        async def _loop(self) -> None:
            await super()._loop()
            logger.info("Client disconnected: %s", self.broker_client_id)
            self.broker.clients.pop(self.broker_client_id, None)

        async def _keep_alive_loop(self) -> None:
            while not self.client.writer.is_closing():
                t = time.monotonic() - self.client.last_activity
                if t < self.IDLE_TIMEOUT:
                    await asyncio.sleep(self.IDLE_TIMEOUT - t)
                else:
                    await self.disconnect()

        async def _message(self, msg: RpcMessage) -> None:
            if self.state == self.State.CONNECTED:
                return await self.client.send(await self._message_hello(msg))
            if self.state == self.State.HELLO:
                return await self.client.send(await self._message_login(msg))

            path = msg.shv_path() or ""
            method = msg.method() or ""

            if msg.is_request():
                # Set access granted to the level allowed by user
                assert self.user is not None
                access = self.user.access_level(path, method)
                if access is None:
                    resp = msg.make_response()
                    resp.set_shverror(RpcErrorCode.METHOD_NOT_FOUND, "No access")
                    await self.client.send(resp)
                    return
                if (
                    access is not RpcMethodAccess.ADMIN
                    or msg.rpc_access_grant() is None
                ):
                    msg.set_rpc_access_grant(access)
                # Check if we should delegate it or handle it ourself
                if (cpath := self.broker.peer_on_path(path)) is None:
                    return await super()._message(msg)
                # Propagate to some peer
                msg.set_caller_ids(
                    list(msg.caller_ids() or ()) + [self.broker_client_id]
                )
                msg.set_shv_path(cpath[1])
                await cpath[0].client.send(msg)

            if msg.is_response():
                cids = list(msg.caller_ids() or ())
                if not cids:  # no caller IDs means this is message for us
                    return await super()._message(msg)
                cid = cids.pop()
                msg.set_caller_ids(cids)
                try:
                    peer = self.broker.clients[cid]
                except KeyError:
                    return
                await peer.client.send(msg)

            if msg.is_signal():
                # TODO this might add unnecessary /
                fullpath = "/".join(self.mount_point + [path])
                msg.set_shv_path(fullpath)
                for client in self.broker.clients.values():
                    if client is self:
                        continue
                    if client.get_subscription(fullpath, method) is not None:
                        await client.client.send(msg)

        async def _message_hello(self, msg: RpcMessage) -> RpcMessage:
            resp = msg.make_response()
            if msg.is_request() and msg.method() == "hello":
                self.state = self.State.HELLO
                resp = msg.make_response()
                resp.set_result({"nonce": self._nonce})
            else:
                resp.set_shverror(
                    RpcErrorCode.INVALID_REQUEST, "Only 'hello' is allowed"
                )
            return resp

        async def _message_login(self, msg: RpcMessage) -> RpcMessage:
            resp = msg.make_response()
            if not msg.is_request() or msg.method() != "login":
                resp.set_shverror(
                    RpcErrorCode.INVALID_REQUEST, "Only 'login' is allowed"
                )
                return resp
            params = msg.params()
            resp = msg.make_response()
            if isinstance(params, collections.abc.Mapping):
                self.user = self.broker.config.login(
                    shvget(params, ("login", "user"), "", str),
                    shvget(params, ("login", "password"), "", str),
                    self._nonce,
                    RpcLoginType(
                        shvget(params, ("login", "type"), RpcLoginType.SHA1.value, str)
                    ),
                )
                if self.user is None:
                    resp.set_shverror(
                        RpcErrorCode.METHOD_CALL_EXCEPTION, "Invalid login"
                    )
                    return resp
                mount_point = shvget(
                    params, ("options", "device", "mountPoint"), "", str
                )
                if self.broker.peer_on_path(mount_point) is not None:
                    resp.set_shverror(
                        RpcErrorCode.METHOD_CALL_EXCEPTION,
                        "Mount point already mounted",
                    )
                    return resp
                # TODO rights to mount to this path?
                self.mount_point = mount_point.split("/") if mount_point else []
                self.IDLE_TIMEOUT = float(
                    shvget(
                        params,
                        ("options", "idleWatchDogTimeOut"),
                        self.IDLE_TIMEOUT,
                        int,
                    )
                )
                resp.set_result({"clientId": self.broker_client_id})
                self.state = self.State.LOGIN
            else:
                resp.set_shverror(
                    RpcErrorCode.INVALID_PARAMS, "Invalid type of parameters"
                )
            return resp

        def _ls(self, path: str) -> typing.Iterator[tuple[str, bool | None]]:
            pth = path.split("/") if path else []
            if len(pth) > 0 and pth[0] == ".broker":
                if len(pth) == 1:
                    yield ("app", False)
                    yield ("clients", True)  # There is always at least this client
                    yield ("currentClient", False)
                    return
                if pth[1] in ("app", "currentClient") and len(pth) == 2:
                    return
                if pth[1] == "clients":
                    if len(pth) == 2:
                        for cid, c in self.broker.clients.items():
                            yield (str(cid), True)
                        return
                    try:
                        cid = int(pth[2], 10)
                    except ValueError:
                        pass
                    else:
                        # Deper paths are handled by forwarding to the client
                        assert len(pth) == 3
                        yield ("app", None)
                        return
            else:
                res: set[tuple[str, bool | None]] = set()
                if len(pth) == 0:
                    res.add((".broker", True))
                res |= {
                    (
                        c.mount_point[len(pth)],
                        None if len(c.mount_point) - 1 == len(pth) else True,
                    )
                    for c in self.broker.clients.values()
                    if c.mount_point and c.mount_point[: len(pth)] == pth
                }
                if res:
                    yield from sorted(iter(res))
                    return

            yield from super()._ls(path)

        def _dir(self, path: str) -> typing.Iterator[RpcMethodDesc]:
            yield from super()._dir(path)
            pth = path.split("/") if path else []
            if len(pth) > 0 and pth[0] == ".broker":
                if len(pth) == 2 and pth[1] == "app":
                    yield RpcMethodDesc("ping")
                    yield RpcMethodDesc(
                        "subscribe",
                        RpcMethodSignature.RET_PARAM,
                        access=RpcMethodAccess.READ,
                    )
                    yield RpcMethodDesc(
                        "unsubscribe",
                        RpcMethodSignature.RET_PARAM,
                        access=RpcMethodAccess.READ,
                    )
                    yield RpcMethodDesc(
                        "rejectNotSubscribed",
                        RpcMethodSignature.RET_PARAM,
                        access=RpcMethodAccess.READ,
                    )
                    yield RpcMethodDesc.getter(
                        "mountPoints", access=RpcMethodAccess.SERVICE
                    )
                    return
                if len(pth) == 2 and pth[1] == "currentClient":
                    yield RpcMethodDesc.getter("clientId")
                    yield RpcMethodDesc.getter("mountPoint")
                    return
                if len(pth) == 3 and pth[1] == "clients":
                    yield RpcMethodDesc.getter(
                        "userName", access=RpcMethodAccess.SERVICE
                    )
                    yield RpcMethodDesc.getter(
                        "mountPoint", access=RpcMethodAccess.SERVICE
                    )
                    yield RpcMethodDesc.getter(
                        "subscriptions", access=RpcMethodAccess.SERVICE
                    )
                    yield RpcMethodDesc(
                        "dropClient",
                        RpcMethodSignature.RET_VOID,
                        access=RpcMethodAccess.SERVICE,
                    )
                    yield RpcMethodDesc.getter(
                        "idleTime", access=RpcMethodAccess.SERVICE
                    )
                    yield RpcMethodDesc.getter(
                        "idleTimeMax", access=RpcMethodAccess.SERVICE
                    )
                    return

        async def _method_call(
            self, path: str, method: str, access: RpcMethodAccess, params: SHVType
        ) -> SHVType:
            if path == ".broker/app":
                if method == "ping":
                    return None
                if method == "subscribe":
                    method = shvget(params, "method", "chng", str)
                    path = shvget(params, "path", "", str)
                    # TODO we do not check the access level. The user might subscribe to the
                    # node that it does not have access and thus get values he should not
                    # see.
                    self.subscriptions.add(self.Subscription(path, method))
                    return None
                if method == "unsubscribe":
                    method = shvget(params, "method", "chng", str)
                    path = shvget(params, "path", "", str)
                    try:
                        self.subscriptions.remove(self.Subscription(path, method))
                    except KeyError:
                        return False
                    return True
                if method == "rejectNotSubscribed":
                    method = shvget(params, "method", "chng", str)
                    path = shvget(params, "path", "", str)
                    sub = self.get_subscription(path, method)
                    if sub is not None:
                        self.subscriptions.remove(sub)
                        return True
                    return False
                if method == "mountPoints" and access >= RpcMethodAccess.SERVICE:
                    return {
                        "/".join(c.mount_point): c.broker_client_id
                        for c in self.broker.clients.values()
                        if c.mount_point
                    }
            elif path == ".broker/currentClient":
                if method == "clientId":
                    return self.broker_client_id
                if method == "mountPoint":
                    return "/".join(self.mount_point) if self.mount_point else None
            elif path.startswith(".broker/clients/"):
                cidstr = path[16:]
                try:
                    cid = int(cidstr, 10)
                except ValueError:
                    pass
                else:
                    if (
                        client := self.broker.clients.get(cid, None)
                    ) is not None and access >= RpcMethodAccess.SERVICE:
                        if method == "userName":
                            assert client.user is not None
                            return client.user.name
                        if method == "mountPoint":
                            return (
                                "/".join(self.mount_point) if self.mount_point else None
                            )
                        if method == "subscriptions":
                            return [
                                {"path": s.path, "method": s.method}
                                for s in client.subscriptions
                            ]
                        if method == "dropClient":
                            await client.disconnect()
                            return True
                        if method == "idleTime":
                            return int(
                                (time.monotonic() - client.client.last_activity) * 1000
                            )
                        if method == "idleTimeMax":
                            return int(self.IDLE_TIMEOUT * 1000)
            return await super()._method_call(path, method, access, params)

        def get_subscription(
            self, path: str, method: str
        ) -> typing.Optional["RpcBroker.Client.Subscription"]:
            """Get subscription matching given path and method.

            This is not exact get, it rather implements standard path and method
            lookup. Thus more specific paths are prefered.
            """
            subs = {s.path: s for s in self.subscriptions if s.method == method}
            pth = path.split("/")
            paths = ("/".join(pth[: len(pth) - i]) for i in range(len(pth)))
            return next((subs[path] for path in paths if path in subs), None)

        async def disconnect(self) -> None:
            logger.info("Disconnecting client: %d", self.broker_client_id)
            self.broker.clients.pop(self.broker_client_id, None)
            await super().disconnect()



"""Check implemntation of SimpleClient."""
import dataclasses

import pytest

from shv import (
    RpcLoginType,
    RpcMethodAccess,
    RpcMethodDesc,
    RpcMethodFlags,
    RpcMethodSignature,
    SHVUInt,
    SimpleClient,
    ValueClient,
    shvmeta_eq,
)


@pytest.mark.parametrize(
    "path,method,params,result",
    (
        (".broker/app", "echo", 42, 42),
        ("", "ls", None, [".broker"]),
        (
            "",
            "dir",
            None,
            [
                {
                    "accessGrant": "bws",
                    "flags": SHVUInt(0),
                    "name": "dir",
                    "signature": 3,
                },
                {
                    "accessGrant": "bws",
                    "flags": SHVUInt(0),
                    "name": "ls",
                    "signature": 3,
                },
            ],
        ),
    ),
)
async def test_call(client, path, method, params, result):
    """Check that we can call various methods using blocking call."""
    res = await client.call(path, method, params)
    assert shvmeta_eq(res, result)


@pytest.mark.parametrize(
    "path,result",
    (
        ("", [".broker"]),
        (".broker", ["app", "clients", "currentClient", "etc", "masters", "mounts"]),
        (".broker/app/log", []),
    ),
)
async def test_ls(client, path, result):
    """Verify that we can use ls method."""
    res = await client.ls(path)
    assert res == result


@pytest.mark.parametrize(
    "path,result",
    (
        ("", {".broker": True}),
        (
            ".broker",
            {
                "app": True,
                "clients": True,
                "currentClient": False,
                "etc": True,
                "masters": False,
                "mounts": False,
            },
        ),
        (".broker/app/log", {}),
    ),
)
async def test_ls_with_children(client, path, result):
    """Verify that we can use ls_with_children method."""
    res = await client.ls_with_children(path)
    assert res == result


@pytest.mark.parametrize(
    "path,result",
    (
        ("", ["dir", "ls"]),
        (
            ".broker/app/log",
            [
                "dir",
                "ls",
                "chng",
                "getSendLogAsSignalEnabled",
                "setSendLogAsSignalEnabled",
                "verbosity",
                "setVerbosity",
            ],
        ),
    ),
)
async def test_dir(client, path, result):
    """Verify that we can use dir method."""
    res = await client.dir(path)
    assert res == result


@pytest.mark.parametrize(
    "path,result",
    (
        (
            "",
            [
                RpcMethodDesc("dir", RpcMethodSignature.RET_PARAM),
                RpcMethodDesc("ls", RpcMethodSignature.RET_PARAM),
            ],
        ),
        (
            ".broker/app/log",
            [
                RpcMethodDesc("dir", RpcMethodSignature.RET_PARAM),
                RpcMethodDesc(
                    "ls", RpcMethodSignature.RET_PARAM, access=RpcMethodAccess.READ
                ),
                RpcMethodDesc(
                    "chng",
                    RpcMethodSignature.VOID_PARAM,
                    RpcMethodFlags.SIGNAL,
                    RpcMethodAccess.READ,
                ),
                RpcMethodDesc(
                    "getSendLogAsSignalEnabled",
                    RpcMethodSignature.RET_VOID,
                    RpcMethodFlags.GETTER,
                    RpcMethodAccess.READ,
                ),
                RpcMethodDesc(
                    "setSendLogAsSignalEnabled",
                    RpcMethodSignature.RET_PARAM,
                    RpcMethodFlags.SETTER,
                    RpcMethodAccess.WRITE,
                ),
                RpcMethodDesc(
                    "verbosity",
                    RpcMethodSignature.RET_VOID,
                    RpcMethodFlags.GETTER,
                    RpcMethodAccess.READ,
                ),
                RpcMethodDesc(
                    "setVerbosity",
                    RpcMethodSignature.RET_PARAM,
                    RpcMethodFlags.SETTER,
                    RpcMethodAccess.COMMAND,
                ),
            ],
        ),
    ),
)
async def test_dir_details(client, path, result):
    """Verify that we can use dir method."""
    res = await client.dir_details(path)
    assert res == result


async def test_sha_login(shvbroker, url):
    """Check that we can login with sha1 password.

    Commonly we login with plain password in tests here but we need to check
    ability to login with SHA1 hashed password as well.
    """
    nurl = dataclasses.replace(
        url,
        password="57a261a7bcb9e6cf1db80df501cdd89cee82957e",
        login_type=RpcLoginType.SHA1,
    )
    client = await SimpleClient.connect(nurl)
    assert await client.ls("") == [".broker"]
    await client.disconnect()


async def test_value_client(example_device, url):
    """Check that we correctly cache values in value client."""
    client = await ValueClient.connect(url)
    await client.subscribe("test/device/track")

    assert len(client) == 0
    assert not list(iter(client))
    with pytest.raises(KeyError):
        _ = client["test/device/track/1"]

    await client.get_snapshot("test/device/track")
    assert client["test/device/track/1"] == [0]
    assert len(client) == 8
    assert list(iter(client)) == [f"test/device/track/{i}" for i in range(1, 9)]

    await client.prop_set("test/device/track/1", [1, 2])
    assert client["test/device/track/1"] == [1, 2]



"""Types used in SHV RPC method description."""
import collections.abc
import dataclasses
import enum
import functools

from .value import SHVType


class RpcMethodSignature(enum.IntEnum):
    """Signature of the SHV RPC method."""

    VOID_VOID = 0
    VOID_PARAM = 1
    RET_VOID = 2
    RET_PARAM = 3


class RpcMethodFlags(enum.IntFlag):
    """Flags assigned to the SHV RPC methods."""

    SIGNAL = 1 << 0
    GETTER = 1 << 1
    SETTER = 1 << 2
    LARGE_RESULT_HINT = 1 << 3


class RpcMethodAccess(enum.IntEnum):
    """Method access level."""

    BROWSE = enum.auto()
    READ = enum.auto()
    WRITE = enum.auto()
    COMMAND = enum.auto()
    CONFIG = enum.auto()
    SERVICE = enum.auto()
    SUPER_SERVICE = enum.auto()
    DEVEL = enum.auto()
    ADMIN = enum.auto()

    @classmethod
    @functools.cache
    def strmap(cls) -> dict[str, "RpcMethodAccess"]:
        return {
            "bws": cls.BROWSE,
            "rd": cls.READ,
            "wr": cls.WRITE,
            "cmd": cls.COMMAND,
            "cfg": cls.CONFIG,
            "srv": cls.SERVICE,
            "ssrv": cls.SUPER_SERVICE,
            "dev": cls.DEVEL,
            "su": cls.ADMIN,
        }

    @classmethod
    @functools.cache
    def strrmap(cls) -> dict[int, str]:
        return {v.value: k for k, v in cls.strmap().items()}

    def tostr(self) -> str:
        """Convert to string representation."""
        return self.strrmap().get(self.value, "bws")

    @classmethod
    def fromstr(cls, access: str) -> "RpcMethodAccess":
        """Convert to string representation."""
        return cls.strmap().get(access, cls.BROWSE)


@dataclasses.dataclass
class RpcMethodDesc:
    """Description of the SHV RPC method.

    This is implemented as :func:`dataclasses.dataclass`.

    :param name: Name of the method.
    :param signature: Calling signature for this method.
    :param flags: Flags assigned to the method.
    :param access: Minimal granted access level for this method.
    :param description: Short description of the method.
    """

    name: str
    signature: RpcMethodSignature = RpcMethodSignature.VOID_VOID
    flags: RpcMethodFlags = RpcMethodFlags(0)
    access: RpcMethodAccess = RpcMethodAccess.BROWSE
    description: str = ""

    def tomap(self) -> SHVType:
        """Convert method description to SHV Map."""
        return {
            "name": self.name,
            "signature": self.signature,
            "flags": self.flags,
            "accessGrant": RpcMethodAccess.tostr(self.access),
            **({"description": self.description} if self.description else {}),
        }

    @classmethod
    def frommap(cls, desc: SHVType) -> "RpcMethodDesc":
        """Convert SHV method description to ."""
        if not isinstance(desc, collections.abc.Mapping):
            raise ValueError(f"Not valid method description: {repr(desc)}")
        name = desc.get("name", None)
        signature = desc.get("signature", None)
        flags = desc.get("flags", None)
        access = desc.get("accessGrant", None)
        if not isinstance(name, str):
            raise ValueError(f"Invalid method name format: {repr(name)}")
        return cls(
            name=name,
            signature=RpcMethodSignature(
                signature if isinstance(signature, int) else 0
            ),
            flags=RpcMethodFlags(flags if isinstance(flags, int) else 0),
            access=RpcMethodAccess.fromstr(
                access if isinstance(access, str) else "bws"
            ),
        )

    @classmethod
    def getter(
        cls,
        name: str = "get",
        access: RpcMethodAccess = RpcMethodAccess.READ,
        description: str = "",
    ) -> "RpcMethodDesc":
        """New getter method description.

        :param name: Name of the method.
        :param access: Minimal granted access level for this getter.
        :param description: Short description of the value.
        """
        return cls(
            name,
            RpcMethodSignature.RET_VOID,
            RpcMethodFlags.GETTER,
            access,
            description,
        )

    @classmethod
    def setter(
        cls,
        name: str = "set",
        access: RpcMethodAccess = RpcMethodAccess.WRITE,
        description: str = "",
    ) -> "RpcMethodDesc":
        """New setter method description.

        :param name: Name of the method.
        :param access: Minimal granted access level for this setter.
        :param description: Short description of the value.
        """
        return cls(
            name,
            RpcMethodSignature.VOID_PARAM,
            RpcMethodFlags.SETTER,
            access,
            description,
        )

    @classmethod
    def signal(
        cls,
        name: str = "chng",
        access: RpcMethodAccess = RpcMethodAccess.WRITE,
        description: str = "",
    ) -> "RpcMethodDesc":
        """New signal method description.

        :param name: Name of the method.
        :param access: Minimal granted access level for this setter.
        :param description: Short description of the value.
        """
        return cls(
            name,
            RpcMethodSignature.RET_VOID,
            RpcMethodFlags.SIGNAL,
            access,
            description,
        )



"""RPC server that waits for clients connection."""
import asyncio
import collections.abc
import logging
import typing

from .rpcclient import RpcClient
from .rpcurl import RpcProtocol, RpcUrl

logger = logging.getLogger(__name__)


async def create_rpc_server(
    client_connected_cb: typing.Callable[
        [RpcClient], None | collections.abc.Awaitable[None]
    ],
    url: RpcUrl,
) -> asyncio.Server:
    """Create server listening on given URL."""

    async def client_connect(
        reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        client = RpcClient(reader, writer)
        res = client_connected_cb(client)
        if isinstance(res, collections.abc.Awaitable):
            await res

    if url.protocol is RpcProtocol.TCP:
        server = await asyncio.start_server(
            client_connect, host=url.host, port=url.port
        )
    elif url.protocol is RpcProtocol.LOCAL_SOCKET:
        server = await asyncio.start_unix_server(client_connect, path=url.host)
    else:
        raise RuntimeError(f"Invalid protocol: {url.protocol}")
    return server



"""RPC connection, that includes client and specific server connection."""
from __future__ import annotations

import asyncio
import io
import logging
import time
import typing

from .chainpack import ChainPack, ChainPackReader, ChainPackWriter
from .cpon import Cpon, CponReader
from .rpcmessage import RpcMessage
from .rpcurl import RpcProtocol

logger = logging.getLogger(__name__)


class RpcClient:
    """RPC connection to some SHV peer.

    You most likely want to use :func:`connect` class methods instead of
    initializing this class directly.

    :param reader: Reader for the connection to the SHV RPC server.
    :param writer: Writer for the connection to the SHV RPC server.
    """

    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ):
        self.last_activity = time.monotonic()
        """Last activity on this connection (read or write of the message)."""
        self.reader = reader
        self.writer = writer
        self._read_data = bytearray(0)

    @classmethod
    async def connect(
        cls,
        host: str | None,
        port: int = 3755,
        protocol: RpcProtocol = RpcProtocol.TCP,
    ) -> RpcClient:
        """Connect to the SHV RPC server.

        :param host: IP/Hostname (TCP) or socket path (LOCKAL_SOCKET)
        :param port: Port (TCP) to connect to
        :param protocol: Protocol used to connect to the server
        """
        if host is None:
            if protocol == RpcProtocol.TCP:
                host = "localhost"
            elif protocol == RpcProtocol.LOCAL_SOCKET:
                host = "shv.sock"
            else:
                raise RuntimeError(f"Invalid protocol: {protocol}")
        logger.debug("Connecting to: %s:%d", host, port)
        if protocol == RpcProtocol.TCP:
            reader, writer = await asyncio.open_connection(host, port)
        elif protocol == RpcProtocol.LOCAL_SOCKET:
            reader, writer = await asyncio.open_unix_connection(host)
        client = cls(reader, writer)
        logger.debug("%s CONNECTED", str(protocol))
        return client

    async def send(self, msg: RpcMessage) -> None:
        """Send the given SHV RPC Message.

        :param msg: Message to be sent
        """
        logger.debug("<== SND: %s", msg.to_string())

        data = msg.to_chainpack()
        writer = ChainPackWriter(self.writer)
        writer.write_uint_data(len(data) + 1)
        self.writer.write(ChainPack.ProtocolType.to_bytes(1, "big"))
        self.writer.write(data)
        await self.writer.drain()
        self.last_activity = time.monotonic()

    def _get_rpc_msg(self):
        if len(self._read_data) < 6:
            return None
        try:
            rd = ChainPackReader(io.BytesIO(self._read_data))
            size = rd.read_uint_data()
            packet_len = size + rd.bytes_cnt
        except EOFError:
            return None
        if packet_len > len(self._read_data):
            return None
        proto = rd.read_raw(1)
        if len(proto) == 1 and proto[0] == Cpon.ProtocolType:
            rd = CponReader(rd)
        rpc_val = rd.read()
        self._read_data = self._read_data[packet_len:]
        return RpcMessage(rpc_val)

    async def receive(self, raise_error: bool = True) -> typing.Optional[RpcMessage]:
        """Read next received RPC message or wait for next to be received.

        :param raise_error: If RpcError should be raised or not.
        :return: Next RPC message is returned or `None` in case of EOF.
        :raise RpcError: When mesasge is error and ``raise_error`` is `True`.
        """
        while not self.reader.at_eof():
            msg = self._get_rpc_msg()
            self.last_activity = time.monotonic()
            if msg:
                logger.debug("==> REC: %s", msg.to_string())
                if raise_error and msg.is_error():
                    raise msg.rpc_error()
                return msg
            data = await self.reader.read(1024)
            self._read_data += data
        return None

    async def disconnect(self):
        """Close the connection."""
        self.writer.close()
        await self.writer.wait_closed()
